# Tutoriel de samurai pour des plasmas

Il s'agit d'une liste de cas tests proposés par Louis Reboul, pour l'instant en 1D.

## Mise en place

### Environnement conda

Il faut avoir installé [conda](https://mamba.readthedocs.io/en/latest/installation/mamba-installation.html) ou mamba ou micromamba. Ensuite il faut créer un environnement contenant les dépendances dont on aura besoin :

* `cxx-compiler` pour avoir un compilateur C++ compatible avec les dépendances
* `cmake` et `make` pour compiler le projet
* `samurai` car on souhaite utiliser [samurai](https://github.com/hpc-maths/samurai)
* `petsc` et `pkg-config` (pour configurer PETSc) s'il nous prenait l'envie d'utiliser des schémas implicites

cela peut se faire via :

```sh
conda create --name tuto-samurai
conda install conda-forge::cmake conda-forge::make conda-forge::cxx-compiler conda-forge::samurai conda-forge::petsc conda-forge::pkg-config conda-forge::h5py conda-forge::numpy conda-forge::python conda-forge::matplotlib
```

ou via un fichier `environnement.yml`

```yml
name: tuto-samurai
channels:
  - conda-forge
dependencies:
  - cmake
  - make
  - cxx-compiler
  - samurai
  - petsc
  - pkg-config
  - h5py
  - numpy
  - python
  - matplotlib
```

et la commande

```sh
conda env create -f environment.yml
```

### Compilation et exécution

La compilation des exemples se fait avec `cmake`. Nous allons compiler et exécuter les codes dans un dossier `build` qui ne servira qu'à ça, pour cela on peut utiliser la commande suivante :

```sh
cmake . -B build
```

Il suffit ensuite d'aller dans le dossier `build/sources` pour compiler un exemple de ce tutoriel, par exemple pour le premier exemple :

```sh
cd build/sources
make tp_1d && ./tp_1d
```

Il sera possible de visualiser la sortie de ce programme avec le script `visu_1d.py` situé à la racine du projet :

```sh
python visu_1d.py build/sources/tp_1d_data --field u
```

## Équation de transport (TP)

> [`sources/tp_1d.cpp`](sources/tp_1d.cpp)

On souhaite résoudre une équation de transport :

$$
  \partial_t u + a \partial_x u = 0
$$

avec $x\in [x_\text{min}, x_\text{max}]$, $L=x_\text{max} - x_\text{min}$.

On peut synthétiser les différents cas sur cette équation dans le tableau suivant.

| Cas | Condition initiale (CI)                            | Condition Limite (CL)                     |
|-----|----------------------------------------------------|-------------------------------------------|
| 1   | $u(0, x) = \sin(\frac{2\pi k x}{L})$               | périodique (pér)                          |
| 2   | $u(0, x) = \sin(\frac{2\pi k x}{L})$               | Dirichlet  (Dir)/Extrapolation (Ext)      |
| 3   | $u(0, x) = \Pi_{[x_\text{min}, x_\text{min}+L/2]}$ | périodique (pér)                          |
| 4   | $u(0, x) = \Pi_{[x_\text{min}, x_\text{min}+L/2]}$ | Dirichlet  (Dir)/Extrapolation (Ext)      |

Les cas 1 et 2 sont avec une condition initiale régulière, les cas 3 et 4 avec un problème de Riemann.


### Définir un opérateur

La bibliothèque samurai propose déjà plusieurs [schémas volumes finis](https://hpc-math-samurai.readthedocs.io/en/latest/reference/finite_volume_schemes.html), dont des schémas de [convection](https://hpc-math-samurai.readthedocs.io/en/latest/reference/finite_volume_schemes.html#id2) (*up-wind* ou WENO5).

Pour cela il faut construire tout d'abord un vecteur vitesse puis l'opérateur de convection :

```cpp
  using field_t = decltype(u_n);                 // need type of field
  samurai::VelocityVector<dim> a_vector = { a }; // a is already define in simulation parameters section
  auto conv = samurai::make_convection_upwind<>( a_vector );
```

Puis on peut utiliser cet opérateur sur différents *fields*, par exemple pour une méthode RK(4,4) on pourrait définir :

```cpp
  u_1 = u_n - dt * conv(u_n);
  u_2 = u_n - 0.5 * dt * conv(u_1);
  u_3 = u_n - dt * conv(u_2);
  u_np1 = -0.33*u_n + 0.33*u_1 + 0.66*u_2 + 0.33*u_3 - 0.16*dt*conv(u_3); 
```

> Il faut évidemment avoir pris soin de définir `u_1`, `u_2`, `u_3` avant, et penser à effectuer des `u_i.resize()` si on souhaite avoir un maillage adaptatif.

### Conditions limites

Pour les conditions limites, si celles-ci sont périodiques, elles se font au niveau de la construction du maillage, en précisant dans quelles dimensions le maillage est périodique avec un tableau (ici appelé `periodic`) par défaut le maillage n'est pas périodique et cet argument est facultatif. Pour les cas 1 et 3 il est donc nécessaire de mettre ce tableau à `true` :

```cpp
  std::array<bool, dim> periodic = {true};
```

Pour d'autres conditions limites, samurai propose tout une interface pour définir des [conditions aux bords](https://hpc-math-samurai.readthedocs.io/en/latest/reference/bc.html), en particulier des conditions de Neumann ou de Dirichlet. Pour cela il faut ajouter des conditions avec la fonction `samurai::make_bc()`, par exemple pour des conditions de Neumann à $0$ d'ordre 1 :

```cpp
  samurai::make_bc<samurai::Neumann<1>>(u_n, 0.);
```


## Flux de Rusanov

> L'équation de transport est le seul cas qui sera traité par des schémas déjà présents dans samurai, la suite utilisera des outils en cours de développement pour définir un schéma volumes-finis tel que Rusanov ou HLL.

L'implémentation proposé par Loïc pour les flux de Rusanov s'appuie sur la définition de flux dans le manuscrit de Louis p127 eq.2.133

$$
  \mathcal{F}^\text{Rus}(u_L, u_R) = \frac{f(u_L) + f(u_R)}{2} - \frac{\lambda_\text{max}}{2}(u_R - u_L)
$$

L'utilisateur doit définir la fonction $f : u \mapsto f(u)$, une direction (dans le cas 1D ce sera toujours `{1}`, il s'agit du vecteur normal à la face pour savoir s'il s'agit d'une face en $x$ ou en $y$), ainsi qu'une fonction donnant la valeur de $\lambda_\text{max}$ à partir de $u_R$ et $u_L$ : $\lambda_\text{max} (u_L, u_R)\mapsto \lambda_\text{max}(u_L, u_R)$. L'utilisateur construit le foncteur de flux à partir de `samurai::Rusanov<with_portions>()`. Ainsi on peut définir un schéma volumes finis en 1D comme suit :

```cpp
  constexpr bool with_portions = false; // define if we want compute flux on finest level
  auto operator = samurai::Rusanov<with_portions>(
    u_n,
    []( const auto& u )
    {
      return f(u); // define function f
    },
    {1}, // direction of flux
    []( const auto& u_l, const auto& u_r)
    {
      return lambda(u_l, u_r); // define estimation of maximal eigenvalue
    }
  );
```

Cet opérateur s'utilise différemment et ne s'appelle que sur système `u_n` :

```
  u_np1 = u_n + dt*operator();
```

où un système est simplement un ensemble de *fields* :

```cpp
  auto u_n = samurai::System(a, b, c);
```


## Équation des ondes à deux variables (WE)

> [`sources/we_1d.cpp`](sources/we_1d.cpp)
>
> manuscrit de Louis : p82 eq.2.26

$$
  \begin{cases}
    \partial_t E + a\partial_x F = 0 \\
    \partial_t F + a\partial_x E = 0
  \end{cases}
$$

avec $x\in [x_\text{min}, x_\text{max}]$, $L=x_\text{max} - x_\text{min}$.


| Cas | CI sur E                  | CI sur F            | CL sur E | CL sur F   |
|-----|---------------------------|---------------------|----------|------------|
| 1   | Régulière                 | Régulière           | Dir      | Neu        |
| 2   | Régulière                 | Régulière           | Neu      | Dir        |
| 3   | $E(0, x) = \Pi_{[\cdot]}$ | $F(0, x) = F_0 = 0$ | Neu      | Neu        |
| 4   | $E(0, x) = E_0$           | $F(0, x) = F_0 = 0$ | Neu      | Dir ($0<F(x_{max})=-F(x_{min})$) |

* Pour le cas 4, on prendra $E_0=1$ et $F(x_{max}) = 1$
* Une variation du cas 4 : imposer $\frac{F}{E}=cst$ sur le bord

Il est également possible de poser $u = E+F$ et $v = E-F$ pour obtenir les équations suivantes :

$$
  \begin{cases}
    \partial_t u + a\partial_x u = 0 \\
    \partial_t v - a\partial_x v = 0
  \end{cases}
$$


## Équation hyperbolique de la chaleur (HHE)

> [`sources/hhe_1d.cpp`](sources/hhe_1d.cpp)
>
> manuscrit de Louis : p147 eq.3.5 et p150 eq.3.18

$$
  \begin{cases}
    \partial_t E + \frac{a}{\varepsilon}\partial_x F = 0 \\
    \partial_t F + \frac{a}{\varepsilon}\partial_x E = -\frac{\sigma}{\varepsilon^2} F 
  \end{cases}
$$

avec $x\in [x_\text{min}, x_\text{max}]$, $L=x_\text{max} - x_\text{min}$, $a=1$.


| Cas | CI sur E                  | CI sur F                  | CL sur E | CL sur F |
|-----|---------------------------|---------------------------|----------|----------|
| 1   | Régulière                 | Régulière                 | Dir      | Neu      |
| 2   | $E(0, x) = \Pi_{[\cdot]}$ | $F(0, x) = \Pi_{[\cdot]}$ | Neu      | Neu      |


## Équation des ondes avec ionisation (IWE)

> [`sources/iwe_1d.cpp`](sources/iwe_1d.cpp)

$$
  \begin{cases}
    \partial_t E + a\partial_x F = \nu_{iz}E \\
    \partial_t F + a\partial_x E = 0
  \end{cases}
$$

avec $\nu_{iz}$ tel que $\partial_t\int_{\Omega}E\textrm{d}x=0$ i.e.

$$
\nu_{iz}=\frac{a\left(F(x_{max})-F(x_{min})\right)}{\int_{\Omega}E\textrm{d}x}
$$

| Cas | CI sur E        | CI sur F            | CL sur E | CL sur F                         |
|-----|-----------------|---------------------|----------|----------------------------------|
| 1   | $E(0, x) = E_0$ | $F(0, x) = F_0 = 0$ | Neu      | Dir ($0<F(x_{max})=-F(x_{min})$) |

(mêmes conditions que 4 dans WE ainsi que sa variante)

version plus stable (avec collisions) :

$$
  \begin{cases}
    \partial_t E + a\partial_x F = \nu_{iz}E \\
    \partial_t F + a\partial_x E = -\sigma F
  \end{cases}
$$


## Euler isotherme (Eiso)

> [`sources/eiso_1d.cpp`](sources/eiso_1d.cpp)
>
> manuscrit de Louis :  p132 eq.2.140

$$
  \begin{cases}
    \partial_t n + \partial_x (nu) = 0 \\
    \partial_t (nu) + \partial_x (nu^2 + nc^2) = 0
  \end{cases}
$$

avec $c^2 = \frac{k_BT}{m}$ ($m$ ?)

| Cas | CI sur n | CI sur nu    | CL sur n | CL sur nu   |
|-----|----------|--------------|----------|-------------|
| 1   | $\Pi$    | $(nu)_0 = 0$ | Neu      | Neu         |

