#pragma once

#include "system.hpp"

#include <functional>
#include <samurai/cell.hpp>
#include <samurai/field.hpp>
#include <type_traits>
#include <utility>

using namespace samurai::literals;

namespace xt
{
    template <>
    struct temporary_type<double>
    {
        using type = double;
    };

    template <class mesh_t, class value_t, std::size_t size, bool SOA>
    struct temporary_type<samurai::Field<mesh_t, value_t, size, SOA>>
    {
        using type = samurai::Field<mesh_t, value_t, size, SOA>;
    };
}

namespace samurai
{

    template <bool with_portion, class Unknowns, class D>
    class Base_scheme : public system_expression<Base_scheme<with_portion, Unknowns, D>>
    {
      public:

        using derived_type               = D;
        static constexpr std::size_t dim = Unknowns::dim;
        using mesh_t                     = typename Unknowns::mesh_t;
        using interval_t                 = typename mesh_t::interval_t;
        using mesh_id_t                  = typename mesh_t::mesh_id_t;
        using cell_t                     = Cell<dim, typename mesh_t::interval_t>;
        using value_t                    = std::conditional_t<with_portion,
                                                              decltype(std::declval<Unknowns>()(cell_t(), std::function<std::array<interval_t, dim>(std::size_t)>{})),
                                                              decltype(std::declval<Unknowns>()[cell_t()])>;
        using temporary_t                = typename xt::temporary_type<std::decay_t<value_t>>::type;
        using arg_t                      = std::conditional_t<std::is_same<value_t, double>::value, double, const value_t&>;
        using function_t                 = std::function<temporary_t(arg_t)>;
        using direction_t                = xt::xtensor_fixed<int, xt::xshape<dim>>;

        Base_scheme(Unknowns& u, const function_t& f, const direction_t& d)
            : m_u(u)
            , m_f(f)
            , m_direction{d}
            , m_result{copy(u)}
        {
        }

        inline void update()
        {
            m_result.resize();
            m_result.fill(0);
            auto& mesh = m_u.mesh();

            // static_assert(std::is_same<decltype(m_result), void>::value);
            auto kernel_inside = [&](const auto& real_cell, const auto& cells)
            {
                auto& mesh  = m_u.mesh();
                auto& left  = cells[0];
                auto& right = cells[1];
                auto f      = compute_flux(left, right);

                m_result[real_cell[0]] += xt::eval(f / std::pow(real_cell[0].length, dim)
                                                   * std::pow(1. / mesh.cell_length(mesh.max_level()), dim - 1));
                m_result[real_cell[1]] -= xt::eval(f / std::pow(real_cell[1].length, dim)
                                                   * std::pow(1. / mesh.cell_length(mesh.max_level()), dim - 1));
                // m_result[real_cell[0]] += xt::eval(f / std::pow(real_cell[0].length, dim) * std::pow(left.length, dim - 1));
                // m_result[real_cell[1]] -= xt::eval(f / std::pow(real_cell[1].length, dim) * std::pow(left.length, dim - 1));
            };

            auto kernel_bc = [&](const auto&, const auto& cells)
            {
                auto& left  = cells[0];
                auto& right = cells[1];
                double dx   = left.length;
                // FIX: Why we need xt::eval here ?
                auto f = xt::eval(compute_flux(left, right) / dx);
                m_result[left] += f;
                m_result[right] -= f;
            };

            // dx flux
            for_each_interior_interface(mesh, m_direction, kernel_inside);
            // for_each_interior_interface(mesh, m_direction, kernel_bc);

            for_each_boundary_interface(mesh, kernel_bc);
        }

        inline auto& operator()()
        {
            update();
            return m_result;
        }

        template <class Cell>
        inline auto u(const Cell& cell)
        {
            return m_u[cell];
        }

        template <class Cell, class Func>
        inline auto u(const Cell& cell, Func&& compute_ind)
        {
            return m_u(cell, std::forward<Func>(compute_ind));
            // return m_u[cell];
        }

        inline auto f(arg_t u)
        {
            return m_f(u);
        }

        template <class Cell>
        inline auto flux(const Cell& left, const Cell& right)
        {
            if constexpr (with_portion)
            {
                auto ul = u(left,
                            [&](auto delta)
                            {
                                std::array<interval_t, dim> block;
                                for (std::size_t d = 0; d < dim; ++d)
                                {
                                    if (m_direction[d] == 0)
                                    {
                                        block[d] = interval_t{0, 1 << delta};
                                    }
                                    else
                                    {
                                        block[d] = interval_t{(1 << delta) - 1, (1 << delta)};
                                    }
                                }
                                return block;
                            });
                auto ur = u(right,
                            [&](auto delta)
                            {
                                std::array<interval_t, dim> block;
                                for (std::size_t d = 0; d < dim; ++d)
                                {
                                    if (m_direction[d] == 0)
                                    {
                                        block[d] = interval_t{0, 1 << delta};
                                    }
                                    else
                                    {
                                        block[d] = interval_t{0, 1};
                                    }
                                }
                                return block;
                            });
                return static_cast<derived_type*>(this)->return_flux(left, right, ul, ur);
            }
            else
            {
                auto ul = u(left);
                auto ur = u(right);
                return static_cast<derived_type*>(this)->return_flux(left, right, ul, ur);
            }
        }

        template <class Cell>
        auto inline compute_flux(const Cell& left, const Cell& right)
        {
            return static_cast<derived_type*>(this)->flux(left, right);
        }

        inline const auto& arguments() const
        {
            return m_u;
        }

        inline auto& arguments()
        {
            return m_u;
        }

        inline auto& result()
        {
            return m_result;
        }

      private:

        Unknowns& m_u;
        function_t m_f;
        direction_t m_direction;
        xt::temporary_type_t<Unknowns> m_result;
    };

    template <bool with_portion, class Unknowns>
    class Rusanov_impl : public Base_scheme<with_portion, Unknowns, Rusanov_impl<with_portion, Unknowns>>
    {
      public:

        using base_class                 = Base_scheme<with_portion, Unknowns, Rusanov_impl<with_portion, Unknowns>>;
        static constexpr std::size_t dim = base_class::dim;
        using value_t                    = typename base_class::value_t;
        using function_t                 = typename base_class::function_t;
        using lambda_t                   = std::function<double(const value_t&, const value_t&)>;
        using direction_t                = typename base_class::direction_t;
        using base_class::f;
        using base_class::u;

        Rusanov_impl(Unknowns& u, const function_t& f, const direction_t& d, const lambda_t& lambda)
            : base_class(u, f, d)
            , m_lambda(lambda)
        {
        }

        template <class Cell, class T>
        auto return_flux(const Cell&, const Cell&, const T& ul, const T& ur)
        {
            auto c = m_lambda(ul, ur);
            return xt::eval(0.5 * (f(ul) + f(ur)) - 0.5 * c * (ur - ul));
        }

      private:

        lambda_t m_lambda;
    };

    template <bool with_portion, class Unknowns, class function_t, class lambda_t>
    auto Rusanov(Unknowns& u, function_t&& f, const xt::xtensor<int, 1>& d, lambda_t&& lambda)
    {
        return Rusanov_impl<with_portion, Unknowns>(u, std::forward<function_t>(f), d, std::forward<lambda_t>(lambda));
    }

    template <bool with_portion, class Unknowns>
    class HLL_impl : public Base_scheme<with_portion, Unknowns, HLL_impl<with_portion, Unknowns>>
    {
      public:

        using base_class                 = Base_scheme<with_portion, Unknowns, HLL_impl<with_portion, Unknowns>>;
        static constexpr std::size_t dim = base_class::dim;
        using cell_t                     = typename base_class::cell_t;
        using value_t                    = typename base_class::value_t;
        using function_t                 = typename base_class::function_t;
        using direction_t                = typename base_class::direction_t;

        using roe_t     = std::function<double(const cell_t&, const cell_t&, const value_t&, const value_t&)>;
        using scaling_t = std::function<double(const cell_t&, const cell_t&)>;
        using base_class::f;
        using base_class::u;

        HLL_impl(Unknowns& u, const function_t& f, const direction_t& d, const roe_t& roe, double c, double scaling = 1.)
            : base_class(u, f, d)
            , m_roe(roe)
            , m_c(c)
            , m_scaling(
                  [scaling](const auto&, const auto&)
                  {
                      return scaling;
                  })
        {
        }

        HLL_impl(Unknowns& u, const function_t& f, const direction_t& d, const roe_t& roe, double c, const scaling_t& scaling)
            : base_class(u, f, d)
            , m_roe(roe)
            , m_c(c)
            , m_scaling(scaling)
        {
        }

        template <class Cell, class T>
        auto return_flux(const Cell& left, const Cell& right, const T& ul, const T& ur)
        {
            auto roe = m_roe(left, right, ul, ur);
            auto bp  = std::max(0., roe + m_c);
            auto bm  = std::min(0., roe - m_c);

            return xt::eval(0.5 * (f(ul) + f(ur))
                            + m_scaling(left, right) / (bp - bm) * (0.5 * (bp + bm) * (f(ul) - f(ur)) + bp * bm * (ur - ul)));
            // return xt::eval(1. / (bp - bm) * (bp * f(ul) - bm * f(ur) + bp * bm * (ur - ul)));
        }

      private:

        roe_t m_roe;
        double m_c;
        scaling_t m_scaling;
    };

    template <bool with_portion, class Unknowns, class function_t, class roe_t>
    auto HLL(Unknowns& u, function_t&& f, const xt::xtensor<int, 1>& d, roe_t&& roe, double c)
    {
        return HLL_impl<with_portion, Unknowns>(u, std::forward<function_t>(f), d, std::forward<roe_t>(roe), c);
    }

    template <bool with_portion, class Unknowns, class function_t, class roe_t, class scaling_t>
    auto HLL(Unknowns& u, function_t&& f, const xt::xtensor<int, 1>& d, roe_t&& roe, double c, scaling_t&& scaling)
    {
        return HLL_impl<with_portion, Unknowns>(u, std::forward<function_t>(f), d, std::forward<roe_t>(roe), c, std::forward<scaling_t>(scaling));
    }
}
