#pragma once

#include "system_expression.hpp"
#include "tuple_container.hpp"
#include "utils.hpp"
#include <samurai/reconstruction.hpp>
#include <tuple>
#include <xtensor/xscalar.hpp>

namespace samurai
{

    template <class Field, class... Fields>
    class System_impl;

    template <std::size_t i, class T>
    typename std::enable_if<!is_system_expression<T>::value, const T&>::type get_entry(const T& t)
    {
        // std::cout << "const Other " << i << std::endl;
        return t;
    }

    template <std::size_t i, class T>
    typename std::enable_if<!is_system_expression<T>::value, T&>::type get_entry(T& t)
    {
        // std::cout << "Other " << i << std::endl;
        return t;
    }

    template <std::size_t I, class... Types>
    auto& get_entry(std::tuple<Types...>& t)
    {
        // std::cout << "tuple " << I << std::endl;
        return get_entry<I>(std::get<I>(t));
    }

    template <std::size_t I, class... Types>
    const auto& get_entry(const std::tuple<Types...>& t)
    {
        // std::cout << "const tuple " << I << std::endl;
        return get_entry<I>(std::get<I>(t));
    }

    template <std::size_t i, class F, class Tuple, std::size_t... I>
    auto myapply_impl(F&&, Tuple&& t, std::index_sequence<I...>)
    {
        return make_field_function<std::decay_t<F>>(get_entry<i>(std::get<I>(t))...);
    }

    template <std::size_t i, class F, class Tuple>
    auto myapply(F&& f, Tuple&& t)
    {
        return myapply_impl<i>(std::forward<F>(f), std::forward<Tuple>(t), std::make_index_sequence<std::tuple_size_v<std::decay_t<Tuple>>>());
    }

    template <std::size_t i, class F, class... CT>
    auto get_entry(const system_function<F, CT...>& e)
    {
        // std::cout << "const system_function " << i << std::endl;
        return myapply<i>(e.function(), e.arguments());
    }

    template <std::size_t i, class E>
    auto& get_entry(system_expression<E>& e)
    {
        // std::cout << "system_expression " << i << std::endl;
        return get_entry<i>(e.derived_cast().arguments());
    }

    template <std::size_t i, class E>
    const auto& get_entry(const system_expression<E>& e)
    {
        // std::cout << "const system_expression " << i << std::endl;
        return get_entry<i>(e.derived_cast().arguments());
    }

    template <std::size_t I, class System, class Expression>
    void assign_on_entry(System& sys, Expression& e)
    {
        get_entry<I>(sys) = get_entry<I>(e);
    }

    template <class System, class Expression, std::size_t... I>
    void assign_impl(System& sys, Expression& e, std::index_sequence<I...>)
    {
        (assign_on_entry<I>(sys, e), ...);
    }

    template <class System, class Expression>
    void assign(System& sys, Expression& e)
    {
        assign_impl(sys, e, std::make_index_sequence<System::size>());
    }

    template <class... Fields>
    auto System(Fields&&... fields)
    {
        return System_impl<std::decay_t<Fields>&...>(std::forward<Fields>(fields)...);
    }

    template <class Field, std::size_t size>
    auto get_portion(std::integral_constant<std::size_t, size>,
                     const Field& t,
                     std::size_t level,
                     const typename Field::interval_t& i,
                     typename Field::interval_t::value_t j,
                     std::size_t delta,
                     const typename Field::interval_t& ii,
                     const typename Field::interval_t& jj)
    {
        return xt::eval(xt::view(portion(t, level, i, j, delta, ii, jj), 0));
    }

    template <class Field, std::size_t size>
    auto get_portion(std::integral_constant<std::size_t, size>,
                     const Field& t,
                     std::size_t level,
                     const typename Field::interval_t& i,
                     typename Field::interval_t::value_t j,
                     typename Field::interval_t::value_t k,
                     std::size_t delta,
                     const typename Field::interval_t& ii,
                     const typename Field::interval_t& jj,
                     const typename Field::interval_t& kk)
    {
        return xt::eval(xt::view(portion(t, level, i, j, k, delta, ii, jj, kk), 0));
    }

    template <class Field>
    auto get_portion(std::integral_constant<std::size_t, 1>,
                     const Field& t,
                     std::size_t level,
                     const typename Field::interval_t& i,
                     typename Field::interval_t::value_t j,
                     std::size_t delta,
                     const typename Field::interval_t& ii,
                     const typename Field::interval_t& jj)
    {
        return portion(t, level, i, j, delta, ii, jj)[0];
    }

    template <class Field>
    auto get_portion(std::integral_constant<std::size_t, 1>,
                     const Field& t,
                     std::size_t level,
                     const typename Field::interval_t& i,
                     typename Field::interval_t::value_t j,
                     typename Field::interval_t::value_t k,
                     std::size_t delta,
                     const typename Field::interval_t& ii,
                     const typename Field::interval_t& jj,
                     const typename Field::interval_t& kk)
    {
        return portion(t, level, i, j, k, delta, ii, jj, kk)[0];
    }

    template <class Field, class Cell>
    auto get_portion(const Field& t, const Cell& cell, std::size_t delta, const std::array<typename Field::interval_t, 2>& block)
    {
        return get_portion(std::integral_constant<std::size_t, Field::size>{},
                           t,
                           cell.level,
                           typename Field::interval_t{cell.indices[0], cell.indices[0] + 1},
                           cell.indices[1],
                           delta,
                           block[0],
                           block[1]);
    }

    template <class Field, class Cell>
    auto get_portion(const Field& t, const Cell& cell, std::size_t delta, const std::array<typename Field::interval_t, 3>& block)
    {
        return get_portion(std::integral_constant<std::size_t, Field::size>{},
                           t,
                           cell.level,
                           typename Field::interval_t{cell.indices[0], cell.indices[0] + 1},
                           cell.indices[1],
                           cell.indices[2],
                           delta,
                           block[0],
                           block[1],
                           block[2]);
    }

    template <class Field, class... Fields>
    class System_impl : public system_expression<System_impl<Field, Fields...>>
    {
      public:

        using temporary_type             = System_impl<std::decay_t<Field>, std::decay_t<Fields>...>;
        using tuple_t                    = std::tuple<Field, Fields...>;
        using concrete_t                 = tuple_t;
        using mesh_t                     = typename std::decay_t<Field>::mesh_t;
        using interval_t                 = typename mesh_t::interval_t;
        static constexpr std::size_t dim = mesh_t::dim;
        using cell_t                     = Cell<dim, typename mesh_t::interval_t>;
        using value_type = std::common_type_t<typename std::decay_t<Field>::value_type, typename std::decay_t<Fields>::value_type...>;
        static constexpr std::size_t size = sizeof...(Fields) + 1;

        template <class... CTA>
        System_impl(CTA&&... fields)
            : m_fields(std::forward<CTA>(fields)...)
        {
        }

        void resize()
        {
            for_each(m_fields,
                     [](auto& f)
                     {
                         f.resize();
                     });
        }

        template <std::size_t I>
        auto& operator[](index<I>)
        {
            static_assert(I < size);
            return std::get<I>(m_fields);
        }

        template <std::size_t I>
        const auto& operator[](index<I>) const
        {
            static_assert(I < size);
            return std::get<I>(m_fields);
        }

        decltype(auto) operator[](const cell_t& cell)
        {
            return std::apply(
                [&](auto&... t)
                {
                    // return std::make_tuple(t[cell]...);
                    return Tuple_container<true>(t[cell]...);
                },
                m_fields);
        }

        template <class Func>
        decltype(auto) operator()(const cell_t& cell, Func&& f)
        {
            if constexpr (dim == 1)
            {
                auto& m     = mesh();
                auto deltal = m.max_level() - cell.level;
                auto block  = f(deltal);
                return std::apply(
                    [&](auto&... t)
                    {
                        return Tuple_container<false>(
                            samurai::portion(t, cell.level, interval_t{cell.indices[0], cell.indices[0] + 1}, deltal, block[0])[0]...);
                    },
                    m_fields);
            }
            else if constexpr (dim > 1)
            {
                auto& m     = mesh();
                auto deltal = m.max_level() - cell.level;
                auto block  = f(deltal);
                return std::apply(
                    [&](auto&... t)
                    {
                        return Tuple_container<false>(get_portion(t, cell, deltal, block)...);
                    },
                    m_fields);
            }
        }

        void fill(value_type v)
        {
            for_each(m_fields,
                     [v](auto& f)
                     {
                         f.fill(v);
                     });
        }

        const auto& mesh() const
        {
            return std::get<0>(m_fields).mesh();
        }

        auto& mesh()
        {
            return std::get<0>(m_fields).mesh();
        }

        template <class E>
        void operator=(const system_expression<E>& e)
        {
            assign(*this, e.derived_cast());
        }

        const auto& arguments() const
        {
            return m_fields;
        }

        auto& arguments()
        {
            return m_fields;
        }

        void to_stream(std::ostream& os) const
        {
            os << "System" << std::endl;
            std::apply(
                [&os](auto&... e)
                {
                    (os << ... << e) << std::endl;
                },
                m_fields);
        }

      private:

        concrete_t m_fields;
    };

    template <class... Fields>
    inline std::ostream& operator<<(std::ostream& out, const System_impl<Fields...>& sys)
    {
        sys.to_stream(out);
        return out;
    }

    template <class... Fields>
    auto copy(System_impl<Fields...>& sys)
    {
        using temporary_type = typename System_impl<Fields...>::temporary_type;
        return std::apply(
            [](auto&... t)
            {
                return temporary_type(t...);
            },
            sys.arguments());
    }

    template <class T>
    auto copy(T& t)
    {
        return t;
    }

}
