#include <algorithm>
#include <cmath>
#include <filesystem>
#include <numbers>

#include <samurai/field.hpp>
#include <samurai/hdf5.hpp>
#include <samurai/mr/adapt.hpp>
#include <samurai/mr/mesh.hpp>
#include <samurai/schemes/fv.hpp>

/*
Solving: d_t u + a d_x u = 0
*/

template <class Field>
void save(std::filesystem::path const& path, std::string const& filename, Field& u, std::string const& suffix = "")
{
    auto mesh   = u.mesh();
    auto level_ = samurai::make_field<std::size_t, 1>("level", mesh);
    u.name()    = "u";

    if (!std::filesystem::exists(path))
    {
        std::filesystem::create_directory(path);
    }

    samurai::for_each_cell(mesh,
                           [&](auto& cell)
                           {
                               level_[cell] = cell.level;
                           });

    samurai::save(path, fmt::format("{}{}", filename, suffix), mesh, u, level_);
}

int main()
{
    // simulation parameters --------------------------------------------------
    constexpr std::size_t dim = 1;

    constexpr double left_box  = 0.;
    constexpr double right_box = 1.;
    constexpr double L         = right_box - left_box; // domain length

    constexpr double t_ini = 0.;
    constexpr double t_end = 1.;

    constexpr double a   = 1.0; // advection velocity
    constexpr double k   = 1.;  // wave number
    constexpr double cfl = 0.9;

    // output parameters
    std::string const dirname = "tp_1d_data";
    fs::path path             = std::filesystem::path(dirname);
    std::string filename      = "u";
    fs::create_directories(path);

    // multiresolution parameters
    std::size_t min_level = 2;
    std::size_t max_level = 8;
    double mr_epsilon     = 1e-3; // Threshold used by multiresolution
    double mr_regularity  = 1.;   // Regularity guess for multiresolution

    // init solution ----------------------------------------------------------
    // define mesh
    using config_t = samurai::MRConfig<dim, 3>;
    using box_t    = samurai::Box<double, dim>;
    using point_t  = typename box_t::point_t;
    point_t box_corner1, box_corner2;
    box_corner1.fill(left_box);
    box_corner2.fill(right_box);
    box_t box(box_corner1, box_corner2);
    std::array<bool, dim> periodic;
    periodic.fill(true);
    samurai::MRMesh<config_t> mesh{box, min_level, max_level, periodic};

    // define field
    auto u_n   = samurai::make_field<1>("u", mesh);
    auto u_np1 = u_n; // temporary field for time scheme
    u_n.fill(0);

    //// sinus initial condition
    samurai::for_each_cell(mesh,
                           [&](auto& cell)
                           {
                               auto x    = cell.center()[0];
                               u_n[cell] = std::sin(2. * std::numbers::pi * k * x / L);
                           });

    //// créneau
    // samurai::for_each_cell(mesh, [&](auto &cell) {
    //   auto x = cell.center()[0];
    //   if (x < 0.25 * L) {
    //     u_n[cell] = 1.0;
    //   }
    // });

    // boundary condition on u_n
    // samurai::make_bc<samurai::Dirichlet<1>>(u_n, 0.);

    // define operator --------------------------------------------------------
    samurai::VelocityVector<dim> a_vector = {a};
    auto conv                             = samurai::make_convection_upwind<decltype(u_n)>(a_vector);

    // adaptation operator ----------------------------------------------------
    auto mr_adaptation = samurai::make_MRAdapt(u_n);
    mr_adaptation(mr_epsilon, mr_regularity);

    // time loop  -------------------------------------------------------------
    double t_n         = t_ini;
    double dt          = cfl * mesh.cell_length(max_level) / a;
    std::size_t n_save = 0;

    samurai::update_ghost_mr(u_n);

    while (t_n < t_end)
    {
        mr_adaptation(mr_epsilon, mr_regularity);
        samurai::update_ghost_mr(u_n);

        u_np1.resize();
        u_np1.fill(0.);

        u_np1 = u_n - dt * conv(u_n);
        t_n += dt;

        std::swap(u_n.array(), u_np1.array());

        save(path, filename, u_n, fmt::format("_ite_{}", n_save++));
    }
}
