#include <algorithm>
#include <cmath>
#include <filesystem>
#include <numbers>

#include <samurai/field.hpp>
#include <samurai/hdf5.hpp>
#include <samurai/interface.hpp>
#include <samurai/mr/adapt.hpp>
#include <samurai/mr/mesh.hpp>
#include <samurai/schemes/fv.hpp>

#include "fv_experimental.hpp"

/*
Solving:
| d_t E + a d_x F = 0
| d_t F + a d_x E = 0
*/

template <class Field_1, class Field_2>
void save(std::filesystem::path const& path, std::string const& filename, Field_1& u, Field_2& v, std::string const& suffix = "")
{
    auto mesh   = u.mesh();
    auto level_ = samurai::make_field<std::size_t, 1>("level", mesh);

    if (!std::filesystem::exists(path))
    {
        std::filesystem::create_directory(path);
    }

    samurai::for_each_cell(mesh,
                           [&](auto& cell)
                           {
                               level_[cell] = cell.level;
                           });

    samurai::save(path, fmt::format("{}{}", filename, suffix), mesh, u, v, level_);
}

int main()
{
    // simulation parameters --------------------------------------------------
    constexpr std::size_t dim = 1;

    constexpr double left_box  = 0.;
    constexpr double right_box = 1.;
    constexpr double L         = right_box - left_box; // domain length

    constexpr double t_ini = 0.;
    constexpr double t_end = 1.;

    constexpr double a                  = 1.0; // advection velocity
    [[maybe_unused]] constexpr double k = 1.;  // wave number
    constexpr double cfl                = 0.9;

    // output parameters
    std::string const dirname = "we_1d_data";
    fs::path path             = std::filesystem::path(dirname);
    fs::create_directories(path);

    // multiresolution parameters
    std::size_t min_level = 6;
    std::size_t max_level = 6;
    double mr_epsilon     = 1e-3; // Threshold used by multiresolution
    double mr_regularity  = 1.;   // Regularity guess for multiresolution

    // init solution ----------------------------------------------------------
    // define mesh
    using config_t = samurai::MRConfig<dim, 3>;
    using box_t    = samurai::Box<double, dim>;
    using point_t  = typename box_t::point_t;
    point_t box_corner1, box_corner2;
    box_corner1.fill(left_box);
    box_corner2.fill(right_box);
    box_t box(box_corner1, box_corner2);
    samurai::MRMesh<config_t> mesh{box, min_level, max_level};

    // define field
    auto E_n = samurai::make_field<1>("E", mesh);
    auto F_n = samurai::make_field<1>("F", mesh);
    E_n.fill(0);
    F_n.fill(0);

    // define system
    auto U_n   = samurai::System(E_n, F_n);
    auto U_np1 = samurai::copy(U_n);

    //// sinus initial condition
    samurai::for_each_cell(mesh,
                           [&](auto& cell)
                           {
                               auto x    = cell.center()[0];
                               E_n[cell] = 0.;
                               F_n[cell] = 0.;
                           });

    // boundary conditions on E and F

    // define operator --------------------------------------------------------

    // adaptation operator ----------------------------------------------------
    auto mr_adaptation = samurai::make_MRAdapt(E_n, F_n);
    mr_adaptation(mr_epsilon, mr_regularity);

    // time loop  -------------------------------------------------------------
    double t_n         = t_ini;
    double dt          = cfl * mesh.cell_length(max_level) / a;
    std::size_t n_save = 0;

    save(path, "EF", E_n, F_n, fmt::format("_ite_{}", n_save++));

    samurai::update_ghost_mr(E_n);
    samurai::update_ghost_mr(F_n);

    while (t_n < t_end)
    {
        // mr_adaptation(mr_epsilon, mr_regularity);
        samurai::update_ghost_mr(E_n);
        samurai::update_ghost_mr(F_n);

        U_np1.resize();
        U_np1.fill(0.);

        // time scheme
        // U_np1 = ...

        t_n += dt;

        // swap solution in U_np1 with solution fields
        std::swap(E_n.array(), U_np1[0_c].array());
        std::swap(F_n.array(), U_np1[1_c].array());

        save(path, "EF", E_n, F_n, fmt::format("_ite_{}", n_save++));
    }
}
