#!/usr/bin/env python

import os
import glob
import sys
import argparse

import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np
import h5py


def import_h5(filename_f5):
    file = h5py.File(filename_f5, 'r')

    mesh = file['mesh']
    points = mesh['points']
    connectivity = mesh['connectivity']

    squares = np.array([
        points[connectivity[i]][:, :-1]
        for i in range(connectivity.shape[0])
    ])
    dim = squares.shape[1]//2

    if dim == 1:
        centers = np.array([np.mean(pts, axis=0)[0] for pts in squares])

        index = np.argsort(centers)

        squares = squares[index]
        centers = centers[index]

        fields = {
            key: mesh['fields'][key][:][index]
            for key in mesh['fields'].keys()
        }
    else:
        centers = np.array([np.mean(pts, axis=0) for pts in squares])

        fields = {
            key: mesh['fields'][key][:]
            for key in mesh['fields'].keys()
        }

    file.close()

    return squares, centers, fields


class h5_data:
    def __init__(self, filename_h5):
        self.squares, self.centers, self.fields = import_h5(filename_h5)

    @property
    def x(self):
        return self.centers[:]

    @property
    def dim(self):
        return self.squares.shape[1]//2

    def __getitem__(self, field):
        return self.fields[field][:]

    def keys(self):
        return self.fields.keys()


def progressbar(it, label="", size=60, out=sys.stdout):
    """
      sources : https://stackoverflow.com/questions/3160699/python-progress-bar#answer-34482761
    """
    count = len(it)
    len_count = int(np.log10(count))
    # start = time.time() # time estimate start

    def show(j):
        x = int(size*j/count)
        # time estimate calculation and string
        # remaining = ((time.time() - start) / j) * (count - j)
        # mins, sec = divmod(remaining, 60) # limited to minutes
        # time_str = f"{int(mins):02}:{sec:03.1f}"
        # print(f"{label}[{u'█'*x}{('.'*(size-x))}] {j}/{count} Est wait {time_str}", end='\r', file=out, flush=True)
        print(f"{label} │{u'█'*x}{('░'*(size-x))}│ {int(j): {len_count+1}d}/{count}",
              end=' \r', file=out, flush=True)

    show(0.1)  # avoid div/0

    for i, item in enumerate(it):
        yield item
        show(i+1)

    print("\n", flush=True, file=out)


parser = argparse.ArgumentParser(
    description="visualization of 1D hdf5 data with animation")
parser.add_argument('directory', type=str,
                    help="directory with hdf5 files finished by `_nnn.h5` (sorted list of nnn makes the animation)")
parser.add_argument('--fields', nargs='+', type=str,
                    required=False, help='list of fields to plot')
parser.add_argument('--ymin', action="extend", nargs='+',
                    required=False, type=float, help='specify a ymin value for rescaling of plots (one by field)')
parser.add_argument('--ymax', action="extend", nargs='+',
                    required=False, type=float, help='specify a ymax value for rescaling of plots (one by field)')
args = parser.parse_args()

# assume we want read all h5 files in this directory and they all finish with `***_n.h5` with `n` the number of iteration
# 1. list all "*.h5" files in given directory
# 2. filter data which doesn't finish with `_n.h5` pattern (`n` is not a digit)
# 3. sort data by `n`
files = sorted(
    filter(lambda f: f.split("/")[-1].split("_")[-1][:-3].isdigit(),
           glob.glob(os.path.join(args.directory, "*.h5"))),
    key=lambda filename: int(filename.split("/")[-1].split("_")[-1][:-3])
)

data = [h5_data(filename)
        for filename in progressbar(files, label="load data")]

fields = args.fields if args.fields is not None else data[0].keys()


def get_ylim_i(i, args):
    y_min = None
    y_max = None

    if args.ymin:
        y_min = args.ymin[i]

    if args.ymax:
        y_max = args.ymax[i]

    return (y_min, y_max)


ylim = [get_ylim_i(i, args) for i in range(len(fields))]

# plot


def plot_1d(i, ax, data, field):
    ax.set_ylim(*ylim[i])
    return ax.plot(data.x, data[field])[0]


def plot_2d(i, ax, data, field):
    return ax.tricontourf(data.x, data.y, data[field], levels=200, cmap="RdBu_r")


plot_func = [plot_1d, plot_2d]

# update


def update_plot_1d(p, data, field):
    p.set_xdata(data.x)
    p.set_ydata(data[field])


def update_plot_2d(p, data, field):
    # p.axes.clear()
    p.axes.tricontourf(data.x, data.y, data[field], levels=200, cmap="RdBu_r")


update_plot_func = [update_plot_1d, update_plot_2d]

fig, axs = plt.subplots(len(fields))
fig.suptitle(f"iteration 0")

for field, ax in zip(fields, axs):
    ax.set_title(field)

plots = [plot_func[data[0].dim - 1](i, ax, data[0], field)
         for i, (field, ax) in enumerate(zip(fields, axs))]


def update(frame):
    fig.suptitle(f"iteration {frame}")
    for field, plot in zip(fields, plots):
        update_plot_func[data[frame].dim - 1](plot, data[frame], field)
        # plot.set_xdata(data[frame].x)
        # plot.set_ydata(data[frame][field])

    return plots[0]


N_frames = len(data)

ani = animation.FuncAnimation(
    fig=fig, func=update, frames=N_frames, interval=10)

plt.show()
